"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DevCycleClient = exports.isDeferredOptions = void 0;
const Variable_1 = require("./Variable");
const Request_1 = require("./Request");
const CacheStore_1 = __importDefault(require("./CacheStore"));
const DefaultStorage_1 = __importDefault(require("./DefaultStorage"));
const User_1 = require("./User");
const EventQueue_1 = require("./EventQueue");
const utils_1 = require("./utils");
const EventEmitter_1 = require("./EventEmitter");
const types_1 = require("@devcycle/types");
const ConfigRequestConsolidator_1 = require("./ConfigRequestConsolidator");
const logger_1 = require("./logger");
const StreamingConnection_1 = require("./StreamingConnection");
const isDeferredOptions = (arg) => {
    return !!arg && 'deferInitialization' in arg;
};
exports.isDeferredOptions = isDeferredOptions;
class DevCycleClient {
    get isInitialized() {
        return this._isInitialized;
    }
    constructor(sdkKey, userOrOptions, optionsArg = {}) {
        this._isInitialized = false;
        this.userSaved = false;
        this._closing = false;
        this.isConfigCached = false;
        this.initializeTriggered = false;
        /**
         * Logic to initialize the client with the appropriate user and configuration data by making requests to DevCycle
         * and loading from local storage. This either happens immediately on client initialization, or when the user is
         * first identified (in deferred mode)
         * @param initialUser
         */
        this.clientInitialization = async (initialUser) => {
            var _a, _b;
            if (this.initializeTriggered || this._closing) {
                return this;
            }
            this.initializeTriggered = true;
            const storedAnonymousId = await this.store.loadAnonUserId();
            this.user = new User_1.DVCPopulatedUser(initialUser, this.options, undefined, storedAnonymousId);
            await this.getConfigCache(this.user);
            // set up requestConsolidator and hook up callback methods
            this.requestConsolidator = new ConfigRequestConsolidator_1.ConfigRequestConsolidator((user, extraParams) => (0, Request_1.getConfigJson)(this.sdkKey, user, this.logger, this.options, extraParams), (config, user) => this.handleConfigReceived(config, user, Date.now()), this.user);
            try {
                await this.requestConsolidator.queue(this.user);
            }
            catch (err) {
                this.eventEmitter.emitInitialized(false);
                this.eventEmitter.emitError(err);
                return this;
            }
            finally {
                this.resolveOnInitialized(this);
                this.logger.info('Client initialized');
            }
            this.eventEmitter.emitInitialized(true);
            if (initialUser.isAnonymous) {
                this.store.saveAnonUserId(this.user.user_id);
            }
            else {
                this.store.removeAnonUserId();
            }
            if ((_b = (_a = this.config) === null || _a === void 0 ? void 0 : _a.sse) === null || _b === void 0 ? void 0 : _b.url) {
                if (!this.options.disableRealtimeUpdates) {
                    this.streamingConnection = new StreamingConnection_1.StreamingConnection(this.config.sse.url, this.onSSEMessage.bind(this), this.logger);
                }
                else {
                    this.logger.info('Disabling Realtime Updates based on Initialization parameter');
                }
            }
            return this;
        };
        let user;
        let options = optionsArg;
        if ((0, exports.isDeferredOptions)(userOrOptions)) {
            options = userOrOptions;
        }
        else {
            user = userOrOptions;
        }
        this.logger =
            options.logger || (0, logger_1.dvcDefaultLogger)({ level: options.logLevel });
        this.store = new CacheStore_1.default(options.storage || new DefaultStorage_1.default(), this.logger);
        this.options = options;
        this.sdkKey = sdkKey;
        this.variableDefaultMap = {};
        this.eventQueue = new EventQueue_1.EventQueue(sdkKey, this, options);
        this.eventEmitter = new EventEmitter_1.EventEmitter();
        this.registerVisibilityChangeHandler();
        this.onInitialized = new Promise((resolve, reject) => {
            this.resolveOnInitialized = (value) => {
                this._isInitialized = true;
                resolve(value);
            };
        });
        if (!this.options.deferInitialization) {
            if (!user) {
                throw new Error('User must be provided to initialize SDK');
            }
            this.clientInitialization(user);
        }
        if (!(options === null || options === void 0 ? void 0 : options.reactNative) && typeof window !== 'undefined') {
            this.windowMessageHandler = (event) => {
                const message = event.data;
                if ((message === null || message === void 0 ? void 0 : message.type) === 'DVC.optIn.saved') {
                    this.refetchConfig(false);
                }
            };
            window.addEventListener('message', this.windowMessageHandler);
            this.windowPageHideHandler = () => {
                this.flushEvents();
            };
            window.addEventListener('pagehide', this.windowPageHideHandler);
        }
    }
    onClientInitialized(onInitialized) {
        if (onInitialized && typeof onInitialized === 'function') {
            this.onInitialized
                .then(() => onInitialized(null, this))
                .catch((err) => onInitialized(err));
            return;
        }
        return this.onInitialized;
    }
    /**
     * Get variable object associated with Features. Use the variable's key to fetch the DVCVariable object.
     * If the user does not receive the feature, the default value is used in the returned DVCVariable object.
     * DVCVariable is returned, which has a `value` property that is used to grab the variable value,
     * and a convenience method to pass in a callback to notify the user when the value has changed from the server.
     *
     * @param key
     * @param defaultValue
     */
    variable(key, defaultValue) {
        var _a, _b;
        if (defaultValue === undefined || defaultValue === null) {
            throw new Error('Default value is a required param');
        }
        // this will throw if type is invalid
        const type = (0, types_1.getVariableTypeFromValue)(defaultValue, key, this.logger, true);
        const defaultValueKey = typeof defaultValue === 'string'
            ? defaultValue
            : JSON.stringify(defaultValue);
        let variable;
        if (this.variableDefaultMap[key] &&
            this.variableDefaultMap[key][defaultValueKey]) {
            variable = this.variableDefaultMap[key][defaultValueKey];
        }
        else {
            const configVariable = (_b = (_a = this.config) === null || _a === void 0 ? void 0 : _a.variables) === null || _b === void 0 ? void 0 : _b[key];
            const data = {
                key,
                defaultValue,
            };
            if (configVariable) {
                if (configVariable.type === type) {
                    data.value = configVariable.value;
                    data.evalReason = configVariable.evalReason;
                }
                else {
                    this.logger.warn(`Type mismatch for variable ${key}. Expected ${type}, got ${configVariable.type}`);
                }
            }
            variable = new Variable_1.DVCVariable(data);
            this.variableDefaultMap[key] = {
                [defaultValueKey]: variable,
                ...this.variableDefaultMap[key],
            };
        }
        this.trackVariableEvaluated(variable);
        this.eventEmitter.emitVariableEvaluated(variable);
        return variable;
    }
    trackVariableEvaluated(variable) {
        var _a, _b;
        if (this.options.disableAutomaticEventLogging)
            return;
        try {
            const variableFromConfig = (_b = (_a = this.config) === null || _a === void 0 ? void 0 : _a.variables) === null || _b === void 0 ? void 0 : _b[variable.key];
            this.eventQueue.queueAggregateEvent({
                type: variable.isDefaulted
                    ? EventQueue_1.EventTypes.variableDefaulted
                    : EventQueue_1.EventTypes.variableEvaluated,
                target: variable.key,
                metaData: {
                    value: variable.value,
                    type: (0, types_1.getVariableTypeFromValue)(variable.defaultValue, variable.key, this.logger),
                    _variable: variableFromConfig === null || variableFromConfig === void 0 ? void 0 : variableFromConfig._id,
                },
            });
        }
        catch (e) {
            this.eventEmitter.emitError(e);
            this.logger.warn(`Error with queueing aggregate events ${e}`);
        }
    }
    /**
     * Get a variable's value associated with a Feature. Use the variable's key to fetch the variable's value.
     * If the user is not segmented into the feature, the default value is returned.
     *
     * @param key
     * @param defaultValue
     */
    variableValue(key, defaultValue) {
        return this.variable(key, defaultValue).value;
    }
    identifyUser(user, callback) {
        const promise = this._identifyUser(user);
        if (callback && typeof callback == 'function') {
            promise
                .then((variables) => callback(null, variables))
                .catch((err) => callback(err, null));
            return;
        }
        return promise;
    }
    async _identifyUser(user) {
        var _a;
        let updatedUser;
        if (this.options.deferInitialization && !this.initializeTriggered) {
            await this.clientInitialization(user);
            return ((_a = this.config) === null || _a === void 0 ? void 0 : _a.variables) || {};
        }
        this.eventQueue.flushEvents();
        try {
            await this.onInitialized;
            const storedAnonymousId = await this.store.loadAnonUserId();
            if (this.user && user.user_id === this.user.user_id) {
                updatedUser = this.user.updateUser(user, this.options);
            }
            else {
                updatedUser = new User_1.DVCPopulatedUser(user, this.options, undefined, storedAnonymousId);
            }
            const config = await this.requestConsolidator.queue(updatedUser);
            if (user.isAnonymous || !user.user_id) {
                this.store.saveAnonUserId(updatedUser.user_id);
            }
            return config.variables || {};
        }
        catch (err) {
            this.eventEmitter.emitError(err);
            throw err;
        }
    }
    resetUser(callback) {
        let oldAnonymousId;
        const anonUser = new User_1.DVCPopulatedUser({ isAnonymous: true }, this.options);
        const promise = new Promise((resolve, reject) => {
            this.eventQueue.flushEvents();
            this.onInitialized
                .then(() => this.store.loadAnonUserId())
                .then((cachedAnonId) => {
                this.store.removeAnonUserId();
                oldAnonymousId = cachedAnonId;
                return;
            })
                .then(() => this.requestConsolidator.queue(anonUser))
                .then((config) => {
                this.store.saveAnonUserId(anonUser.user_id);
                resolve(config.variables || {});
            })
                .catch((e) => {
                this.eventEmitter.emitError(e);
                if (oldAnonymousId) {
                    this.store.saveAnonUserId(oldAnonymousId);
                }
                reject(e);
            });
        });
        if (callback && typeof callback == 'function') {
            promise
                .then((variables) => callback(null, variables))
                .catch((err) => callback(err, null));
            return;
        }
        return promise;
    }
    /**
     * Retrieve data on all Features, Object mapped by feature `key`.
     * Use the `DVCFeature.segmented` value to determine if the user was segmented into a
     * feature's audience.
     */
    allFeatures() {
        var _a;
        return ((_a = this.config) === null || _a === void 0 ? void 0 : _a.features) || {};
    }
    /**
     * Retrieve data on all Variables, Object mapped by variable `key`.
     */
    allVariables() {
        var _a;
        return ((_a = this.config) === null || _a === void 0 ? void 0 : _a.variables) || {};
    }
    subscribe(key, handler) {
        this.eventEmitter.subscribe(key, handler);
    }
    /**
     * Unsubscribe to remove existing event emitter subscription.
     *
     * @param key
     * @param handler
     */
    unsubscribe(key, handler) {
        this.eventEmitter.unsubscribe(key, handler);
    }
    /**
     * Track Event to DevCycle
     *
     * @param event
     */
    track(event) {
        if (this._closing) {
            this.logger.error('Client is closing, cannot track new events.');
            return;
        }
        if (this.options.disableCustomEventLogging)
            return;
        (0, utils_1.checkParamDefined)('type', event.type);
        this.onInitialized.then(() => {
            this.eventQueue.queueEvent(event);
        });
    }
    /**
     * Flush all queued events to DevCycle
     *
     * @param callback
     */
    flushEvents(callback) {
        return this.eventQueue.flushEvents().then(() => callback === null || callback === void 0 ? void 0 : callback());
    }
    /**
     * Close all open connections to DevCycle, flush any pending events and
     * stop any running timers and event handlers. Use to clean up a client instance
     * that is no longer needed.
     */
    async close() {
        var _a;
        this.logger.debug('Closing client');
        this._closing = true;
        if (document && this.pageVisibilityHandler) {
            document.removeEventListener('visibilitychange', this.pageVisibilityHandler);
        }
        if (this.windowMessageHandler) {
            window.removeEventListener('message', this.windowMessageHandler);
        }
        if (this.windowPageHideHandler) {
            window.removeEventListener('pagehide', this.windowPageHideHandler);
        }
        (_a = this.streamingConnection) === null || _a === void 0 ? void 0 : _a.close();
        await this.eventQueue.close();
    }
    /**
     * Reflects whether `close()` has been called on the client instance.
     */
    get closing() {
        return this._closing;
    }
    async refetchConfig(sse, lastModified, etag) {
        await this.onInitialized;
        await this.requestConsolidator.queue(null, { sse, lastModified, etag });
    }
    handleConfigReceived(config, user, dateFetched) {
        var _a;
        const oldConfig = this.config;
        this.config = config;
        this.store.saveConfig(config, user, dateFetched);
        this.isConfigCached = false;
        if (this.user != user || !this.userSaved) {
            this.user = user;
            this.store.saveUser(user);
            if (!this.user.isAnonymous &&
                checkIfEdgeEnabled(config, this.logger, (_a = this.options) === null || _a === void 0 ? void 0 : _a.enableEdgeDB, true)) {
                (0, Request_1.saveEntity)(this.user, this.sdkKey, this.logger, this.options).then((res) => this.logger.info(`Saved response entity! ${res}`));
            }
            this.userSaved = true;
        }
        const oldFeatures = (oldConfig === null || oldConfig === void 0 ? void 0 : oldConfig.features) || {};
        const oldVariables = (oldConfig === null || oldConfig === void 0 ? void 0 : oldConfig.variables) || {};
        this.eventEmitter.emitFeatureUpdates(oldFeatures, config.features);
        this.eventEmitter.emitVariableUpdates(oldVariables, config.variables, this.variableDefaultMap);
        if (!oldConfig || oldConfig.etag !== this.config.etag) {
            this.eventEmitter.emitConfigUpdate(config.variables);
        }
    }
    onSSEMessage(message) {
        var _a, _b;
        try {
            const parsedMessage = JSON.parse(message);
            const messageData = JSON.parse(parsedMessage.data);
            if (!messageData) {
                return;
            }
            if (!messageData.type || messageData.type === 'refetchConfig') {
                if (!((_a = this.config) === null || _a === void 0 ? void 0 : _a.etag) ||
                    messageData.etag !== ((_b = this.config) === null || _b === void 0 ? void 0 : _b.etag)) {
                    this.refetchConfig(true, messageData.lastModified, messageData.etag).catch((e) => {
                        this.logger.warn(`Failed to refetch config ${e}`);
                    });
                }
            }
        }
        catch (e) {
            this.logger.warn(`Streaming Connection: Unparseable message ${e}`);
        }
    }
    registerVisibilityChangeHandler() {
        var _a, _b, _c;
        if (typeof document === 'undefined') {
            return;
        }
        const inactivityDelay = ((_b = (_a = this.config) === null || _a === void 0 ? void 0 : _a.sse) === null || _b === void 0 ? void 0 : _b.inactivityDelay) || 120000;
        this.pageVisibilityHandler = () => {
            var _a, _b, _c;
            if (!((_a = this.config) === null || _a === void 0 ? void 0 : _a.sse)) {
                return;
            }
            else if (document.visibilityState === 'visible') {
                if (!((_b = this.streamingConnection) === null || _b === void 0 ? void 0 : _b.isConnected())) {
                    this.logger.debug('Page became visible, refetching config');
                    this.refetchConfig(false).catch((e) => {
                        this.logger.warn(`Failed to refetch config ${e}`);
                    });
                    (_c = this.streamingConnection) === null || _c === void 0 ? void 0 : _c.reopen();
                }
                window === null || window === void 0 ? void 0 : window.clearTimeout(this.inactivityHandlerId);
            }
            else {
                window === null || window === void 0 ? void 0 : window.clearTimeout(this.inactivityHandlerId);
                this.inactivityHandlerId = window === null || window === void 0 ? void 0 : window.setTimeout(() => {
                    var _a;
                    this.logger.debug('Page is not visible, closing streaming connection');
                    (_a = this.streamingConnection) === null || _a === void 0 ? void 0 : _a.close();
                }, inactivityDelay);
            }
        };
        (_c = document.addEventListener) === null || _c === void 0 ? void 0 : _c.call(document, 'visibilitychange', this.pageVisibilityHandler);
    }
    async getConfigCache(user) {
        if (this.options.disableConfigCache) {
            this.logger.info('Skipping config cache');
            return;
        }
        const cachedConfig = await this.store.loadConfig(user, this.options.configCacheTTL);
        if (cachedConfig) {
            this.config = cachedConfig;
            this.isConfigCached = true;
            this.eventEmitter.emitFeatureUpdates({}, cachedConfig.features);
            this.eventEmitter.emitVariableUpdates({}, cachedConfig.variables, this.variableDefaultMap);
            this.logger.debug('Initialized with a cached config');
        }
    }
}
exports.DevCycleClient = DevCycleClient;
const checkIfEdgeEnabled = (config, logger, enableEdgeDB, logWarning = false) => {
    var _a, _b;
    if ((_b = (_a = config.project.settings) === null || _a === void 0 ? void 0 : _a.edgeDB) === null || _b === void 0 ? void 0 : _b.enabled) {
        return !!enableEdgeDB;
    }
    else {
        if (enableEdgeDB && logWarning) {
            logger.warn('EdgeDB is not enabled for this project. Only using local user data.');
        }
        return false;
    }
};
//# sourceMappingURL=Client.js.map