"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.EventQueue = exports.EventTypes = void 0;
const Request_1 = require("./Request");
const utils_1 = require("./utils");
const chunk_1 = __importDefault(require("lodash/chunk"));
exports.EventTypes = {
    variableEvaluated: 'variableEvaluated',
    variableDefaulted: 'variableDefaulted',
};
class EventQueue {
    constructor(sdkKey, dvcClient, options) {
        var _a, _b;
        this.eventQueueBatchSize = 100;
        this.sdkKey = sdkKey;
        this.client = dvcClient;
        this.eventQueue = [];
        this.aggregateEventMap = {};
        this.options = options;
        const eventFlushIntervalMS = typeof options.eventFlushIntervalMS === 'number'
            ? options.eventFlushIntervalMS
            : 10 * 1000;
        if (eventFlushIntervalMS < 500) {
            throw new Error(`eventFlushIntervalMS: ${eventFlushIntervalMS} must be larger than 500ms`);
        }
        else if (eventFlushIntervalMS > 60 * 1000) {
            throw new Error(`eventFlushIntervalMS: ${eventFlushIntervalMS} must be smaller than 1 minute`);
        }
        this.flushInterval = setInterval(this.flushEvents.bind(this), eventFlushIntervalMS);
        this.flushEventQueueSize = (_a = options === null || options === void 0 ? void 0 : options.flushEventQueueSize) !== null && _a !== void 0 ? _a : 100;
        this.maxEventQueueSize = (_b = options === null || options === void 0 ? void 0 : options.maxEventQueueSize) !== null && _b !== void 0 ? _b : 1000;
        if (this.flushEventQueueSize >= this.maxEventQueueSize) {
            throw new Error(`flushEventQueueSize: ${this.flushEventQueueSize} must be smaller than ` +
                `maxEventQueueSize: ${this.maxEventQueueSize}`);
        }
        else if (this.flushEventQueueSize < 10 ||
            this.flushEventQueueSize > 1000) {
            throw new Error(`flushEventQueueSize: ${this.flushEventQueueSize} must be between 10 and 1000`);
        }
        else if (this.maxEventQueueSize < 100 ||
            this.maxEventQueueSize > 5000) {
            throw new Error(`maxEventQueueSize: ${this.maxEventQueueSize} must be between 100 and 5000`);
        }
    }
    async flushEvents() {
        const user = this.client.user;
        if (!user) {
            this.client.logger.warn('Skipping event flush, user has not been set yet.');
            return;
        }
        const eventsToFlush = [...this.eventQueue];
        const aggregateEventsToFlush = this.eventsFromAggregateEventMap();
        eventsToFlush.push(...aggregateEventsToFlush);
        if (!eventsToFlush.length) {
            return;
        }
        this.client.logger.info(`Flush ${eventsToFlush.length} Events`);
        this.eventQueue = [];
        this.aggregateEventMap = {};
        const eventRequests = (0, chunk_1.default)(eventsToFlush, this.eventQueueBatchSize);
        for (const eventRequest of eventRequests) {
            try {
                const res = await (0, Request_1.publishEvents)(this.sdkKey, this.client.config || null, user, eventRequest, this.client.logger);
                if (res.status === 201) {
                    this.client.logger.info(`DevCycle Flushed ${eventRequest.length} Events.`);
                }
                else if (res.status >= 500 || res.status === 408) {
                    this.client.logger.warn('failed to flush events, retrying events. ' +
                        `Response status: ${res.status}, message: ${res.statusText}`);
                    this.eventQueue.push(...eventRequest);
                }
                else {
                    this.client.logger.error('failed to flush events, dropping events. ' +
                        `Response status: ${res.status}, message: ${res.statusText}`);
                }
            }
            catch (ex) {
                this.client.eventEmitter.emitError(ex);
                this.client.logger.error('failed to flush events due to error, dropping events. ' +
                    `Error message: ${ex === null || ex === void 0 ? void 0 : ex.message}`);
            }
        }
    }
    /**
     * Queue DVCAPIEvent for producing
     */
    queueEvent(event) {
        if (this.checkEventQueueSize()) {
            this.client.logger.warn(`DevCycle: Max event queue size (${this.maxEventQueueSize}) reached, dropping event: ${event}`);
            return;
        }
        this.eventQueue.push(event);
    }
    /**
     * Queue DVCEvent that can be aggregated together, where multiple calls are aggregated
     * by incrementing the 'value' field.
     */
    queueAggregateEvent(event) {
        if (this.checkEventQueueSize()) {
            this.client.logger.warn(`DevCycle: Max event queue size (${this.maxEventQueueSize}) reached, dropping event: ${event}`);
            return;
        }
        (0, utils_1.checkParamDefined)('type', event.type);
        (0, utils_1.checkParamDefined)('target', event.target);
        event.date = Date.now();
        event.value = 1;
        const aggEventType = this.aggregateEventMap[event.type];
        if (!aggEventType) {
            this.aggregateEventMap[event.type] = { [event.target]: event };
        }
        else if (aggEventType[event.target]) {
            aggEventType[event.target].value++;
        }
        else {
            aggEventType[event.target] = event;
        }
    }
    checkEventQueueSize() {
        const aggCount = Object.values(this.aggregateEventMap).reduce((acc, v) => acc + Object.keys(v).length, 0);
        const queueSize = this.eventQueue.length + aggCount;
        if (queueSize >= this.flushEventQueueSize) {
            this.flushEvents();
        }
        return queueSize >= this.maxEventQueueSize;
    }
    /**
     * Turn the Aggregate Event Map into an Array of DVCAPIEvent objects for publishing.
     */
    eventsFromAggregateEventMap() {
        return Object.values(this.aggregateEventMap)
            .map((typeMap) => Object.values(typeMap))
            .flat();
    }
    async close() {
        clearInterval(this.flushInterval);
        await this.flushEvents();
    }
}
exports.EventQueue = EventQueue;
//# sourceMappingURL=EventQueue.js.map